/* $Id: descr_stats.c,v 1.1.1.1 2003/05/22 08:57:32 roca Exp $ */
/*
 *  Copyright (c) 2003 INRIA - All rights reserved
 *  (main author: Vincent Roca - vincent.roca@inrialpes.fr)
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */
/*
 *	This file calculates various statistics on a set values.
 *	(mean, median, variance, standard deviation, confidence interval)
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <math.h>

#ifdef DEBUG
#define TRACE(m)	printf m
#else
#define TRACE(m)
#endif

#define min(a,b)	((a) < (b) ? (a) : (b))
#define max(a,b)	((a) > (b) ? (a) : (b))


#define STEPVAL		0.01	/* size of an interval for histograms */
#define MAXNAMELEN	256	/* max name string length */
#define MAXCMDLEN	2048	/* max name string length */
#define MAXVAL		100000	/* max nb of values */


float	val[MAXVAL];		/* array of values from input file */
int	valnb;			/* nb of values of array */
int	middle;


void
usage (int	argc,
       char	*argv[])
{
	printf("\
    descr_stats - version 1.0 - May 22nd, 2003\n\
    Copyright (c) 2003 INRIA - All rights reserved\n\
    Author:	vincent.roca@inrialpes.fr\n\
    Web site:	http://www.inrialpes.fr/planete/people/roca/\n\
	\n\
    descr_stats comes with ABSOLUTELY NO WARRANTY; This is free software,\n\
    and you are welcome to redistribute it under certain conditions;\n\
    See the GNU General Public License as published by the Free\n\
    Software Foundation, version 2 or later, for more details.\n\
	\n\
    usage: descr_stats col if\n\
	col	column to consider in the input file (first column is 1)\n\
	if	input file\n\
	\n\
    This tool calculates various descriptive statistics on a set of\n\
    samples stored in a text file: mean, median, variance, standard\n\
    deviation, confidence interval around the mean and median. It can\n\
    also produce an histogram of the samples. In that case, be careful\n\
    at the step value used, since this parameter may dramatically\n\
    change the results.\n");
	exit(-1);
}


/*
 * find the confidence interval around the given value
 */
float
confidence_int (float	confidence,
		float	center)
{
	int	i_center;
	int	inf, sup;
	int	need;			/* we need so many values... */
	int	nb;			/* ...but we have this number */
	float	delta_inf;
	float	delta_sup;
	int	found = 0;		/* 1 if i_center has been found */

	for (i_center = 0; i_center < valnb - 1; i_center++) {
		if (val[i_center] < center && val[i_center+1] >= center) {
			found++;
			break;
		}
	}
	if (!found) {
		return -1.0;	/* not found */
	}
	need = (int)floor((double)(valnb * confidence));
	nb = 0;
	inf = sup = i_center;
	while (nb < need) {
		if (inf == 0)
			sup++;
		else if (sup == valnb -1)
			inf--;
		else if ((center - val[inf - 1]) < (val[sup + 1] - center))
			inf--;
		else
			sup++;
		nb++;
	}
	delta_inf = center - val[inf];
	delta_sup = val[sup] - center;
	TRACE(("delta_inf=%f, delta_sup=%f, inf=%d, sup=%d, need=%d\n",
		delta_inf, delta_sup, inf, sup, need));
	return (max(delta_inf, delta_sup));
}


int
main (int	argc,
      char	*argv[])
{
	FILE	*ifp, *ofp;		/* input/output file ptrs */
	char	ifn[MAXNAMELEN];	/* input file name string */
	char	ofn[MAXNAMELEN];	/* output file name string */
	char	demfn[MAXNAMELEN];	/* .dem file name string */
	char	cmd[MAXCMDLEN];		/* command string */
	int	col;			/* colomn number to keep */
	int	i;
	float	sum, mean, median, variance, std_deviation, range;
	char	answer;
	float	upto, step, half_step;
	int	nb_per_int;		/* histograms: sample nb per interval */
	float	int_90_mean;
	float	int_90_median;
	float	int_95_mean;
	float	int_95_median;
	float	int_99_mean;
	float	int_99_median;

	if (argc != 3)
		usage(argc, argv);

	col = atoi(argv[1]);
	strncpy(ifn, argv[2], MAXNAMELEN);

	sprintf(ofn, "/tmp/of%d", getpid());
	TRACE(("\ttemp file is %s\n", ofn));

	/* keep only the appropriate col */
	sprintf(cmd, "cat %s | awk '{ print $%d }' > %s", ifn, col, ofn);
	TRACE(("--> keep col cmd: \"%s\"\n", cmd));
	system(cmd);

	strncpy(ifn, ofn, MAXNAMELEN);
	sprintf(ofn, "/tmp/of%d", getpid()+1);
	TRACE(("\tnew temp file is %s\n", ofn));

	/* sort the file (specify ifn both for input and output file) */
	sprintf(cmd, "sort -n %s > %s", ifn, ofn);
	TRACE(("--> sort cmd: \"%s\"\n", cmd));
	system(cmd);

	strncpy(ifn, ofn, MAXNAMELEN);
	if ((ifp = fopen(ifn, "r")) < 0) {
		perror("open");
		usage(argc, argv);
	}

	/* store values of the input file into our array */
	valnb = 0;
	while (fscanf(ifp, "%f", &val[valnb]) == 1) {
		valnb++;
		if (valnb >= MAXVAL) {
			printf("WARNING: limited to %d values, remaining lines ignored\n", MAXVAL);
			break;
		}
	}
	if (valnb <= 1) {
		printf("ERROR: no value read in %s\n", ifn);
		exit(-1);
	}

	/* now calculate various stats */
	sum = 0.0;
	for (i = 0; i < valnb; i++) {
		sum += val[i];
	}

	mean = sum / valnb;

	if (valnb % 2) {
		/*
		 * odd nb of samples, the simple case.
		 * Be careful, valnb is the number of elements,
		 * while middle is an index (from 0 to valnb-1)!
		 */
		middle = (int)floor((double)valnb / 2.0);
		median = val[middle];
	} else {
		/*
		 * even nb of samples
		 * val[middle] | median | val[middle+1]
		 * in that case consider the mean between these
		 * two values (val[middle] and val[middle+1]) as
		 * the median.
		 * Be careful, valnb is the number of elements,
		 * while middle is an index (from 0 to valnb-1)!
		 */
		middle = (int)floor((double)(valnb - 1) / 2.0);
		median = (val[middle] + val[middle+1]) / 2.0;
	}

	variance = 0.0;
	for (i = 0; i < valnb; i++) {
		variance += (val[i] - mean) * (val[i] - mean);
	}
	variance = variance / (valnb - 1);
	std_deviation = sqrt(variance);

	range = val[valnb -1] - val[0];

	int_90_mean = confidence_int(0.90, mean);
	int_95_mean = confidence_int(0.95, mean);
	int_99_mean = confidence_int(0.99, mean);
	int_90_median = confidence_int(0.90, median);
	int_95_median = confidence_int(0.95, median);
	int_99_median = confidence_int(0.99, median);

	/*
	 * Print results
	 */
	printf(
"	------------------------------------------------------\n\
	nb of samples = %d\n\
	mean = %f\n\
	median = %f\n\
	variance = %f\n\
	standard deviation = %f\n\
	range = %f \t min = %f  max = %f\n",
		valnb, mean, median, variance, std_deviation, range,
		val[0], val[valnb -1]);

	printf("\tconfidence interval around mean %f:\n", mean);
	if (int_90_mean != -1.0) printf("\t\t90: +/- %f\n", int_90_mean); else printf("\t\t90: NOT FOUND\n");
	if (int_95_mean != -1.0) printf("\t\t95: +/- %f\n", int_95_mean); else printf("\t\t95: NOT FOUND\n");
	if (int_99_mean != -1.0) printf("\t\t99: +/- %f\n", int_99_mean); else printf("\t\t99: NOT FOUND\n");

	printf("\tconfidence interval around median %f:\n", median);
	if (int_90_median != -1.0) printf("\t\t90: +/- %f\n", int_90_median); else printf("\t\t90: NOT FOUND\n");
	if (int_95_median != -1.0) printf("\t\t95: +/- %f\n", int_95_median); else printf("\t\t95: NOT FOUND\n");
	if (int_99_median != -1.0) printf("\t\t99: +/- %f\n", int_99_median); else printf("\t\t99: NOT FOUND\n");
	printf("\t------------------------------------------------------\n");
#if 0
	printf(
"	------------------------------------------------------\n\
	nb of samples = %d\n\
	mean = %f\n\
	median = %f\n\
	variance = %f\n\
	standard deviation = %f\n\
	range = %f \t min = %f  max = %f\n\
	confidence interval around mean %f:\n\
		90: +/- %f\n\
		95: +/- %f\n\
		99: +/- %f\n\
	------------------------------------------------------\n",
		valnb, mean, median, variance, std_deviation, range,
		val[0], val[valnb -1],
		mean, int_90_mean, int_95_mean, int_99_mean);
#endif

	/*
	 * Histogram file preparation
	 */
	answer = 'n';
	printf("Continue with histogram (produces a gnuplot .dem format) (y/n)[n] ? ");
	scanf("%c", &answer);
	if (answer != 'y')
		goto cleanup;

	step = STEPVAL;		/* default value */
	half_step = step / 2;
	printf("Enter sampling step (real number > 0.0): ");
	scanf("%f", &step);/* XXX: done this way, the user MUST enter a value */
	if (step <= 0.0) {
		printf("ERROR, step value %f invalid (must be > 0)\n", step);
		goto cleanup;
	}
	sprintf(ofn, "/tmp/histo%d.dat", getpid());
	sprintf(demfn, "/tmp/histo%d.dem", getpid());
	printf("histogram data file is:		%s\n", ofn);
	printf("histogram gnuplot file is:	%s\n", demfn);

	if ((ofp = fopen(ofn, "w")) < 0) {
		perror("open");
		exit(-1);
	}

	for (upto = val[0] + step, i = 0, nb_per_int = 0;
	     i < valnb;
	     i++) {
		if (val[i] < upto) {
			nb_per_int ++;
		} else {
			/* switch to next interval */
			fprintf(ofp, "%f %d\n", upto - half_step, nb_per_int);
			upto = val[i] + step;
			nb_per_int = 1;
		}
	}
	/* don't forget the last one... */
	fprintf(ofp, "%f %d\n", upto - half_step, nb_per_int);

	/* and now prepar the histo.dem file */
	sprintf(cmd, "echo \'
		set title  \"Histogram\"\n\
		set xlabel \"value\"\n\
		set ylabel \"number of samples per %f interval\"\n\
		set autoscale\n\
		set nolabel\n\
		set grid\n\
		plot [] [0:] \"%s\" with impulses\n\
		pause -1 \"Hit return to continue\" \' > %s",
		step, ofn, demfn);
	TRACE(("--> histo.dem creation cmd: \"%s\"\n", cmd));
	system(cmd);

#if 0
	/* and finally launch gnuplot... */
	/*sprintf(cmd, "xterm -e gnuplot -fn 5x8 %s", demfn);*/
	sprintf(cmd, "gnuplot %s", demfn);
	TRACE(("--> gnuplot cmd: \"%s\"\n", cmd));
	system(cmd);
#endif

cleanup:
#ifdef DEBUG
	/* keep temp output files */
#else
	/* remove everything */
	sprintf(cmd, "rm /tmp/of[0-9]*");
	system(cmd);
#endif
	return 0;
}

